open Format
open Typ
open Exp
open Typcontext
open Conv

let rec check ctx e = match e with
  Var(fi,x) -> 
    (try (match lookup_var x ctx with
            Bind(t,_) -> (bind x t (Linear,Nonmodal) empty,Var(fi,x),t)
         |  AbbrevBind(t) -> Support.Error.bug "check")
     with Not_found -> Support.Error.errAt fi ("Unbound variable: " ^ x))
| Con(fi,c) -> (empty,Con(fi,c),TNat)
| SZero(fi) -> (empty,SZero(fi),TArr((Linear,Nonmodal),TNat,TNat))
| SOne(fi) -> (empty,SOne(fi),TArr((Linear,Nonmodal),TNat,TNat))
| Abs(fi,x,_,ty,e) -> 
    let (x',e') = Exp.rename_binding ctx (x,e) in
    let (bodyctx,e,bodyty) = check (bind x' ty unknown_aspect ctx) e' in
    (try (match lookup_var x' bodyctx with 
       Bind(ty,(l,m)) -> let a = if banged ty then (Linear,m) else (l,m) in  
                        (remove_var x' bodyctx,
                              Abs(fi,x,a,ty,e),TArr(a,ty,bodyty))
    | _ -> Support.Error.bug "check")
   with (* Variable not used in body *)
      Not_found -> 
          let m = Nonmodal in 
          let a = if affine or (banged ty) 
                     then (Linear,m) 
                     else (Nonlinear,m) in 
              (bodyctx,Abs(fi,x,a,ty,e),TArr(a,ty,bodyty)))
| App(fi,e1,e2) -> 
    let (ctx1,e1,ty1) = check ctx e1 in
    let (ctx2,e2,ty2) = check ctx e2 in
    (match Conv.find_concrete_supertype ctx ty1 with 
       TArr(a,ty11,ty12) -> 
         if Conv.leq ctx ty2 ty11 
          then (mergeapp a ctx1 ctx2,App(fi,e1,e2),ty12)
          else Support.Error.errfAt fi (function () ->
                print_string "Type of arg does not match domain of function: ";
                Typ.print ty2; print_string "is not a subtype of ";
                Typ.print ty11)
    | TNat -> Support.Error.errfAt fi (function () ->
          print_string "Integer in applied position")
    | _ -> Support.Error.bug "check")        
| Let(fi,x,_,e1,e2) -> 
    let (x,e2) = Exp.rename_binding ctx (x,e2) in
    let (ctx1,e1,ty1) = check ctx e1 in
    let (ctx2,e2,ty2) = check (bind x ty1 unknown_aspect ctx) e2 in
    (try (match lookup_var x ctx2 with 
               Bind(_,(l,m)) -> 
               let l = if banged ty1 then Linear else l in  
               (mergeapp (l,m) ctx1 (remove_var x ctx2),
                 Let(fi,x,(l,m),e1,e2),
                 ty2)
              | _ -> Support.Error.bug "check: let")
     with 
        Not_found -> 
          let m = Nonmodal in 
          let l = if affine or (banged ty1) then Linear else Nonlinear in 
           (mergeapp (l,m) ctx1 (remove_var x ctx2),
            Let(fi,x,(l,m),e1,e2),
            ty2))
| Case(fi,ty,e1,e2,e3,e4) -> 
    let (ctx1,e1,ty1) = check ctx e1 in 
    if not (Conv.leq ctx1 ty1 TNat) then 
       Support.Error.errfAt fi (function () -> 
           print_string "first argument to case has type "; 
           Typ.print ty1;print_string " which is not a subtype of N")
    else 
    let (ctx2,e2,ty2) = check ctx e2 in 
    let (ctx3,e3,ty3) = check ctx e3 in 
    let (ctx4,e4,ty4) = check ctx e4 in 
    if not (leq ctx ty2 ty)
       then Support.Error.errfAt fi (function () ->
        print_string "zero branch of case has type ";Typ.print ty2;
        print_string "which is not a subtype of purported type")
    else if not (leq ctx ty3 (TArr(Typ.unknown_aspect,TNat,ty)))
       then Support.Error.errfAt fi (function () ->
        print_string "even branch of case has type ";Typ.print ty3;
        print_string "which is not a subtype of purported type")
    else if not (leq ctx ty4 (TArr(Typ.unknown_aspect,TNat,ty)))
       then Support.Error.errfAt fi (function () ->
        print_string "odd branch of case has type ";Typ.print ty4;
        print_string "which is not a subtype of purported type")
    else  (mergeapp (Linear,Nonmodal) empty
          (mergeapp (Linear,Nonmodal) ctx1
          (mergecase ctx2 (mergecase ctx3 ctx4))),
           Case(fi,ty,e1,e2,e3,e4),ty)
| Saferec(fi,ty) -> 
    if not (safeallowed ty) then Support.Error.errfAt fi  (function () -> 
       print_string "Safe recursion not allowed at type ";Typ.print ty) 
    else (empty,Saferec(fi,ty),
    TArr((Linear,Modal),TNat,
      TArr((Linear,Nonmodal),ty,
     TArr((Nonlinear,Nonmodal),
       TArr((Linear,Modal),TNat,TArr((Nonlinear,Nonmodal),ty,ty)),ty)))) 
| Linrec(fi,ty) -> 
    if not (linallowed ty) then Support.Error.errfAt fi (function () -> 
       print_string "Linear safe recursion not allowed at type ";Typ.print ty) 
    else (empty,Linrec(fi,ty),
    TArr((Linear,Modal),TNat,TArr((Linear,Nonmodal),ty,
     TArr((Nonlinear,Nonmodal),
       TArr((Linear,Modal),TNat,TArr((Linear,Nonmodal),ty,ty)),ty)))) 
| Coerce(fi,e1,ty2) -> 
    let (ctx1,e1,ty1) = check ctx e1 in
      if Conv.leq ctx ty1 ty2 
         then (ctx1,Coerce(fi,e1,ty2),ty2) 
         else Support.Error.errfAt fi
        (fun () -> print_string "Explicit coercion failed: "; Typ.print ty1;
           print_string "is not a subtype of the expected type ";
           Typ.print ty2)














