(*
 * Type expressions
 *)

type linearity = 
     Linear
  |  Nonlinear

type modality = 
    Modal
  | Nonmodal

type aspect = linearity * modality

(* Nonlinear <: Linear 
 * Modal <: Nonmodal *)

val leq_lin : linearity*linearity -> bool
val leq_mod : modality*modality -> bool
val leq_asp : aspect*aspect -> bool
val unknown_aspect : aspect

type t = 
    TVar of string
  | TNat   
  | TArr of aspect * t * t

(* Types allowed for safe and linear safe recursion *)

val safeallowed : t -> bool
val linallowed : t -> bool

(* Types that are reproducable. If banged A = true then A -o B is
printed as A -> B *)
val banged : t -> bool

(* if affine is true then weakening is allowed in 
linear type, eg fun(x:A)y : A -o B in this case *)
val affine : bool

(* Print a type *) 
val print_bare : t -> unit

(* Print a type, making sure that it gets inset if it's long *) 
val print : t -> unit





