open Format
open Typ

let lin_inf = function 
    (Linear,Linear) -> Linear
  | _ -> Nonlinear

let mod_inf = function
    (Nonmodal,Nonmodal) -> Nonmodal
  | _ -> Modal

let asp_inf = function
    ((l,m),(l',m')) -> (lin_inf(l,l'),mod_inf(m,m'))


type binding = 
    Bind of Typ.t * aspect
  | AbbrevBind of Typ.t

type ctx = binding Support.StringMap.t

let empty = Support.StringMap.empty

let bind x t a ctx = 
  try 
    let _ = Support.StringMap.find x ctx in
    Support.Error.err ("variable " ^ x ^ " is already bound")
  with
    Not_found -> Support.StringMap.add x (Bind(t,a)) ctx

let abbrevbind x t ctx = 
     Support.StringMap.add x (AbbrevBind(t)) (Support.StringMap.remove x ctx)

let dummybind x ctx = bind x Typ.TNat (Nonlinear, Nonmodal) ctx

let lookup_var x ctx = 
  match Support.StringMap.find x ctx with
    Bind(t,a) -> Bind(t,a)
  | _ -> raise Not_found

let remove_var x ctx = 
 Support.StringMap.remove x ctx

let lookup_abbrev x ctx = 
  match Support.StringMap.find x ctx with
      AbbrevBind(t) -> t
    | _ -> raise Not_found

let rec choose_fresh_var x ctx =
  try 
    let b = Support.StringMap.find x ctx in
    match b with
      Bind(_,_)       -> choose_fresh_var (x ^ "'") ctx
    | AbbrevBind(_) -> Support.Error.bug "choose_fresh_var"
  with
    Not_found -> x
  
  

let extract_abbrevs ctx = 
  Support.StringMap.fold 
    (function x -> function b -> function abbrevs ->
      match b with
        AbbrevBind(ty) -> (x,ty)::abbrevs
      | _ -> abbrevs)
    ctx []



let mergeapp a ctx1 ctx2 = 
  Support.StringMap.fold
     (function x -> function b -> function tmp ->  
          (match b with
             Bind(t,(l,m)) -> 
                (try (match lookup_var x tmp with
                     Bind(t,(l',m')) -> 
                       bind x t (Nonlinear,mod_inf(m,m')) (remove_var x tmp) 
                   | _ -> Support.Error.bug "mergelin") with
                 Not_found -> bind x t (l,m) tmp)  
	  | AbbrevBind(t) -> abbrevbind x t (remove_var x tmp))) ctx1 
  (Support.StringMap.fold
     (function x -> function b -> function tmp -> 
          (match b with 
            Bind(t,a') -> bind x t (asp_inf(a,a')) tmp
	  | AbbrevBind(t) -> abbrevbind x t tmp)) ctx2 empty)


let mergecase ctx1 ctx2 = 
  Support.StringMap.fold
     (function x -> function b -> function tmp ->  
          (match b with
             Bind(t,((l,m) as a)) -> 
                (try (match lookup_var x tmp with
                     Bind(t,a') -> 
                      bind x t (asp_inf (a,a')) (remove_var x tmp) 
                                  | _ -> Support.Error.bug "mergecase") with
                 Not_found -> bind x t 
                              ((if affine then l else Nonlinear),m) tmp)  
	  | AbbrevBind(t) -> abbrevbind x t (remove_var x tmp))) ctx1 
(if affine then ctx2 
 else (* we have to nonlinearize variables occurring in ctx2 only *)
 Support.StringMap.fold 
     (function x -> function b -> function tmp ->  
          (match b with
             Bind(t,((l,m) as a)) -> 
                (try (match lookup_var x ctx1 with
                     Bind(t,a') -> bind x t a tmp 
                    | _ -> Support.Error.bug "mergecase") with
                 Not_found -> bind x t  (Nonlinear,m) tmp)  
	  | AbbrevBind(t) -> abbrevbind x t (remove_var x tmp))) ctx2 empty)


       
let nonboxempty ctx =
  Support.StringMap.fold
     (function x -> function b -> function asyet -> 
         asyet & 
         (match b with 
           Bind(t,(l,m)) -> m = Modal 
	 |   _ -> true)) ctx true


let linearity x ctx = match lookup_var x ctx with
  Bind(_,(l,_)) -> l
| _ -> Support.Error.bug "linearity"

let extract_type x ctx = match lookup_var x ctx with
  Bind(t,_) -> t
| _ -> Support.Error.bug "extract_type"


(* For debugging *)


let showctx = 
let lin2string l = if l = Linear then "^" else ":" in
let mod2string m = if m = Modal then "$" else "" in
Support.StringMap.iter
            (function x -> function b  ->
            match b with
              Bind(t,(l,m)) -> 
                print_string ((mod2string m)^x^(lin2string l));Typ.print t;
	    | AbbrevBind(t) -> print_string (x^"=");Typ.print t)





