(* 
 * A typing context is a mapping from 
 *     variable names to types 
 *     type variable names to nothing (just records bindings)
 *     type abbreviation names to types
 *)


type binding = 
    Bind of Typ.t * Typ.aspect
  | AbbrevBind of Typ.t

type ctx

val empty : ctx

(* If a variable to be bound is already present then bind, boxbind,
dummybind generate error messages; abbrevbind overwrites the previous
binding *)

val bind : string -> Typ.t -> Typ.aspect -> ctx -> ctx       
val abbrevbind : string -> Typ.t -> ctx -> ctx  
val dummybind : string -> ctx -> ctx            



val lookup_var : string -> ctx -> binding     (* raises Not_found if unbound *)
val remove_var : string -> ctx -> ctx         

val lookup_abbrev : string -> ctx -> Typ.t    (* raises Not_found if unbound *)

val extract_abbrevs : ctx -> (string * Typ.t) list

val choose_fresh_var : string -> ctx -> string

val nonboxempty : ctx -> bool             (* non box part of arg is empty *)

val mergeapp : Typ.aspect -> ctx -> ctx -> ctx 

(* mergeapp a ctx1 ctx2 returns the principal context of e1 e2 from
the principal contexts of e1 and e2 when e1 is a function of aspect
a. *)

val mergecase : ctx -> ctx -> ctx 

(* mergecase ctx1 ctx2 returns the union of ctx1 ctx2 with the common
variables bound linearly if they are linear in both ctx1 and ctx2 and
modalized if so required in either branch *)

val linearity : string -> ctx -> Typ.linearity
val extract_type : string -> ctx -> Typ.t

(* Printing contexts for debugging *)

val showctx : ctx -> unit


